/*Implementation of 2D FCT through the 1d algorithm of 
  Arguello and Zapata algorithm from paper:

F. Arguello and E.L. Zapata, "Fast cosine transform based on the 
succesive doubling method", Electronics Letters, Vol. 26, No. 19, 
pp.1616-1618, Sept. 1990.

Written by:
 C.A. Christopoulos
Vrije Universiteit Brussel 
VUB - ETRO (IRIS)
Pleinlaan 2
1050 Brussels
Belgium
Tel:	++ 32 - 2 - 6412940
Fax:	++ 32 - 2 - 6412883
E-mail:	chchrist@etro.vub.ac.be
    12 /6/1992
*/


#include <math.h>
#include <stdio.h>
#define PI 3.141592653589

unsigned N,m,N0; double x[512][512],ct[512];

main()
  {
		 double mtest = 1.;
		 int nl, nloops;

		printf("Enter number of points (N=power of 2): ");
		 scanf("%d",&N);
		 printf("Enter no of loops:");
		 scanf("%d",&nloops);
		 mtest=log(N+0.)/log(2.);
                 N0=N;
		 m=mtest;
		/* }*/

  ans_read();
      makecosinetable();
      two_d_throughaz();
  
  ans_write();
 }                      /* end of main */
/*************************************************************/

makecosinetable() /* 1D cosine table for N points */
{
   int k,i,p,s,t;
   
        p=0;
	for(i=1; i<=m; i++){
		s=1<<(i-1); t=1<<i;
		for(k=0; k<N/t; k++){
		    ct[p++]=2*cos(PI*((k<<2)+1)*s/(N<<1));
	        }
         }       
}/* end of cosine table*/

/*************************************************************/
rowscosine()
{
    int i,j,k,p,q,rows;

  for (rows=0; rows<N; rows++) {
     for (i=m-1; i>0; i--) {
	p=1 << (i-1); q=1 << i;
	for(j=(1 << (m-i))-2; j >=0; j--) {
	   for(k=0; k<p-1; k++)
		x[rows][1+(j+1)*q+k] += x[rows][1+j*q+k];
            x[rows][1+((j<<1)+3)*p-1] -= x[rows][1+((j<<1)+1)*p-1];
        }
     }
  }
}
/**************************************************************/
columnscosine()
{
  int i,j,k,p,q,cols;

  for (cols=0; cols<N; cols++) {
     for (i=m-1; i>0; i--) {
        p=1 << (i-1); q=1 << i;
	for(j=(1 << (m-i))-2; j >=0; j--) {
	    for(k=0; k<p-1; k++)
		x[1+(j+1)*q+k][cols] += x[1+j*q+k][cols];
 	    x[1+((j<<1)+3)*p-1][cols] -= x[1+((j<<1)+1)*p-1][cols];
        }
     }
   }
}
/*************************************************************/
 two_d_throughaz() /* 2d using the row-column approach */
  {
    int i,j,k,l,s,t,rows,cols,p;
    register double a;
    
 /* make rows */
    rowsinputmapping(); 
    for (rows=0; rows<N; rows++) {
        p=0;
	for(i=1; i<=m; i++){
		s=1<<(i-1); t=1<<i;
		for(k=0; k<N/t; k++){
		   /* dcos=2*cos(PI*((k<<2)+1)*s/(N<<1));*/
		    for(j=0;j<s;j++){
		   	 l=k+j*N/s; 
			 a=x[rows][l];
			 x[rows][l] += x[rows][l+N/t];
			 x[rows][l+N/t] = (a-x[rows][l+N/t])*ct[p];
		    }
                    p++;
		}
	}
    }
    rowsbitreversal();
    rowscosine();
    for (rows=0; rows<N; rows++) 
        for (cols=1; cols<N; cols++)  
             x[rows][cols] = x[rows][cols] / 2.0;

  /* make columns */
    columnsinputmapping();
    for (cols=0; cols<N; cols++) {
         p=0;
	for(i=1; i<=m; i++){
		 s=1<<(i-1); t=1<<i;
		for(k=0; k<N/t; k++){
		   /* dcos=2*cos(PI*((k<<2)+1)*s/(N<<1));*/
		    for(j=0;j<s;j++){
		   	 l=k+j*N/s; 
			 a=x[l][cols];
			 x[l][cols] += x[l+N/t][cols];
			 x[l+N/t][cols] = (a-x[l+N/t][cols])*ct[p];
		    }
                    p++;
		}
	}
    }
    columnsbitreversal();
    columnscosine();
    for (cols=0; cols<N; cols++) 
        for (rows=1; rows<N; rows++)  
                 x[rows][cols] = x[rows][cols] / 2.0;
	
 
}  /*    end of two_d_throughaz    */
/******************************************************************/
rowsbitreversal() /* 2d bit reversal */
{
    int v1, v2, v3,i,j,k,cols,rows;
    double xt;

    /* revesre rows */
     for (cols =0; cols<N; cols ++) {   
	  v1 = (m+1)/2; 
	  v2 = 1 << v1; 
	  v3 = N-1-v2;
	  j=0;
	  for(i=1; i<=v3; i++){
	    	k= N>>1;
		while(k<=j){
			j=j-k; 
			k=k>>1;
		}
		j +=k;
		if(i<j){
			xt=x[cols][j];
			x[cols][j]=x[cols][i];
			x[cols][i]=xt;
		}
	  }
     }
}
/*****************************************************************/
columnsbitreversal()
 {   
  int v1, v2, v3,i,j,k,cols,rows;
    double xt;
    /* reverse columns */
     for (rows =0; rows<N; rows ++) {   
	  v1 = (m+1)/2; 
	  v2 = 1 << v1; 
	  v3 = N-1-v2;
	  j=0;
	  for(i=1; i<=v3; i++){
	    	k= N>>1;
		while(k<=j){
			j=j-k; 
			k=k>>1;
		}
		j +=k;
		if(i<j){
			xt=x[j][rows];
			x[j][rows]=x[i][rows];
			x[i][rows]=xt;
		}
	  }
     }

} 
/******************************************************************/
columnsinputmapping() /* 2d input mapping */
{
  
	int n1,n2,rows,cols,n; double s[512][512];

  	for(n1=0; n1 <= N; n1++) 
           for (n2=0; n2<N; n2++ ) 
               s[n1][n2] = x[n1][n2];

        for (rows=0; rows<N; rows++) {
  	    for(n=0; n < N/2; n++) {
     		 x[n][rows]     = s[2*n][rows];
     	 	x[N-n-1][rows] = s[2*n+1][rows];
      	    }
        }
} 
/*******************************************************************/
rowsinputmapping()
{	int n1,n2,rows,cols,n; double s[512][512];

  	for(n1=0; n1 <= N; n1++) 
           for (n2=0; n2<N; n2++ ) 
               s[n1][n2] = x[n1][n2];

        for (cols=0; cols<N; cols++) {
  	  for(n=0; n < N/2; n++) {
     		 x[cols][n]     = s[cols][2*n];
     	 	x[cols][N-n-1] = s[cols][2*n+1];
         }
      }
	
 }  
/**********************************************************************/
ans_read()
{
	int rows,cols;
	double value;
	char file_name[50];
	FILE *fi,*fopen();


	printf("Enter name of file to be FCTransformed: ");
	scanf("%s",file_name);
     	if((fi=fopen(file_name,"r"))==NULL) {
		printf("cannot open file  '%s' \n",file_name);
		exit(1);
	}
        for (rows=0; rows<N; rows++) {
            for (cols=0; cols<N; cols++) {
                   if ((fscanf(fi,"%lf",&value)!=EOF))  {
		        x[cols][rows] = value;
                   }
             }
        } 
	fclose(fi);
}/* End of "ans_read" function */

/*-------------------------------------------------------------------*/
ans_write()
{
	int rows,cols;
	FILE *fo,*fopen();

	fo=fopen("DIT_FCT.OUT","w");
        for (rows=0; rows<N; rows++) {
            for (cols=0; cols<N; cols ++) {
               /*printf("%d  %d  %lf\n",cols,rows,x[cols][rows]);*/
               /* fprintf(fo,"%d  %d  %lf\n",cols,rows,x[cols][rows]);*/
            }
        }
 	fclose(fo);

	puts("  ");
	puts("Your transformed output file is:  DIT_FCT.OUT");
} /* end of "ans_write" function */
