/*Implementation of 2D FCT through the 1d algorithm of 
  Chan and Ho from paper:
 
S. C. Chan and K. L. Ho, "A new two-dimensional fast cosine transform 
algorithm", IEEE Trans. on Signal Processing, Vol. 39, No. 2, pp. 481-485,
Feb. 1991.

Written by:
C.A. Christopoulos
Vrije Universiteit Brussel
VUB - ETRO (IRIS)
Pleinlaan 2
1050 Brussels
Belgium
Tel:	++ 32 - 2 - 6412940
Fax:	++ 32 - 2 - 6412883
E-mail:	chchrist@etro.vub.ac.be

   12 /6/1993
*/


#include <math.h>
#include <stdio.h>
#define PI 3.141592653589

unsigned N,m,N0; double x[512][512],ct[512];

main()
  {
		 double mtest = 1.;
		 int nl, nloops;

		printf("Enter number of points (N=power of 2): ");
		 scanf("%d",&N);
		 printf("Enter no of loops:");
		 scanf("%d",&nloops);
		 mtest=log(N+0.)/log(2.);
                 N0=N;
		 m=mtest;
		/* }*/

  ans_read();
     makecosinetable();
      chan();
  
 /* ans_write();*/
 }                      /* end of main */
/*************************************************************/

makecosinetable() /* 1D cosine table for N points */
  {
   int n1,k,j,i,i1,l,n2,rows,cols,p;
   double a,e,c,xt;

         n2 = N;
         p=0;
	 for (k=1; k<m; k++) {
             n1 = n2;
	     n2 = n2 >>1;
	     e = PI/(n1<<1);
	     for (j=0; j<n2; j++) {
	        ct[p++] = cos(((j<<2)+1)*e);
	     }
         }
         ct[p++] = cos (PI/4.0);
  }/* end of cosine table*/

/*************************************************************/
 chan() /* 2d using the row-column approach */
  {
   int n1,k,j,i,i1,l,n2,rows,cols,p;
   double a,e,c,xt;

   
   rowsinputmapping();
    for (rows=0; rows<N; rows++) {
         p=0;
         n2 = N;
	 for (k=1; k<m; k++) {
           n1 = n2;
	   n2 = n2 >>1;
	   e = PI/(n1<<1);
	   for (j=0; j<n2; j++) {
	      c=ct[p++];
	      for (i=j; i<N; i+=n1) {
			 l = i+n2;
			 xt = x[rows][i] - x[rows][l];
			 x[rows][i] = x[rows][i] + x[rows][l];
			 x[rows][l] = 2*xt * c;
		}
	   }
        }
         c = ct[p++];
	for (i=0; i<N; i+=2) {
		i1 = i+1;
		xt = x[rows][i];
		x[rows][i] += x[rows][i1];
		x[rows][i1] = (xt-x[rows][i1])* c;
	 }
    }/* end of for rows */
    rowsbitreversal();
    rowspostadditions();

  columnsinputmapping();	
  for (cols=0; cols<N; cols++) {
         p=0;
         n2 = N;
	 for (k=1; k<m; k++) {
          n1 = n2;
	   n2 = n2 >>1;
	   e = PI/(n1<<1);
	   for (j=0; j<n2; j++) {
	     c=ct[p++];
	      for (i=j; i<N; i+=n1) {
			 l = i+n2;
			 xt = x[i][cols] - x[l][cols];
			 x[i][cols] = x[i][cols] + x[l][cols];
			 x[l][cols] = 2*xt * c;
		}
	   }
	 }
	 c = ct[p++];
	for (i=0; i<N; i+=2) {
		i1 = i+1;
		xt = x[i][cols];
		x[i][cols] += x[i1][cols];
		x[i1][cols] = (xt-x[i1][cols])* c;
	 }
    }/* end of for cols */
    columnsbitreversal();
    columnspostadditions();


  }  /*    end of SR_FCT    */
/******************************************************************/
columnspostadditions()
{
     int step,loops,k,ep,j,i,l,cols,rows;

    for (cols=0; cols<N; cols++) {
        step =N;
	loops = 1;
	for (k=1; k<m; k++)  {
            step = step>>1;
	     ep = step>>1;
	     loops = loops <<1;
	     for (j=0; j<(step>>1); j++) {
	        l=ep;
                 x[l][cols] = x[l][cols]/2;

                for (i=1; i<loops; i++)  {
                     x[l+step][cols] = x[l+step][cols] - x[l][cols];
                   l =l+step;
                }
		 ep +=1;
	     }
	}
    }
 } /* end of rowspostadditions */   
/*****************************************************************/
rowspostadditions()
{
 int step,loops,k,ep,j,i,l,cols,rows;


      /* Postaditions for the columns */
      for (rows=0; rows<N; rows++) {
         step =N;
	 loops = 1;
	   for (k=1; k<m; k++)  {
                step = step>>1;
	        ep = step>>1;
            	loops = loops <<1;
		for (j=0; j<(step>>1); j++) {
	            l=ep;
                    x[rows][l]=x[rows][l]/2;
                     for (i=1; i<loops;i++)  {
                       x[rows][l+step] = x[rows][l+step]-x[rows][l];
                       l =l+step;
	            }
		    ep +=1;
		  }
	     }
       }
        
     
} /* end of columnspostadditions */

/******************************************************************/
rowsbitreversal() /* 2d bit reversal */
{
    int v1, v2, v3,i,j,k,cols,rows;
    double xt;

    /* revesre rows */
     for (cols =0; cols<N; cols ++) {   
	  v1 = (m+1)/2; 
	  v2 = 1 << v1; 
	  v3 = N-1-v2;
	  j=0;
	  for(i=1; i<=v3; i++){
	    	k= N>>1;
		while(k<=j){
			j=j-k; 
			k=k>>1;
		}
		j +=k;
		if(i<j){
			xt=x[cols][j];
			x[cols][j]=x[cols][i];
			x[cols][i]=xt;
		}
	  }
     }
}
/*****************************************************************/
columnsbitreversal()
 {   
  int v1, v2, v3,i,j,k,cols,rows;
    double xt;
    /* reverse columns */
     for (rows =0; rows<N; rows ++) {   
	  v1 = (m+1)/2; 
	  v2 = 1 << v1; 
	  v3 = N-1-v2;
	  j=0;
	  for(i=1; i<=v3; i++){
	    	k= N>>1;
		while(k<=j){
			j=j-k; 
			k=k>>1;
		}
		j +=k;
		if(i<j){
			xt=x[j][rows];
			x[j][rows]=x[i][rows];
			x[i][rows]=xt;
		}
	  }
     }

} 
/******************************************************************/
columnsinputmapping() /* 2d input mapping */
{
  
	int n1,n2,rows,cols,n; double s[512][512];

  	for(n1=0; n1 <= N; n1++) 
           for (n2=0; n2<N; n2++ ) 
               s[n1][n2] = x[n1][n2];

        for (rows=0; rows<N; rows++) {
  	    for(n=0; n < N/2; n++) {
     		 x[n][rows]     = s[2*n][rows];
     	 	x[N-n-1][rows] = s[2*n+1][rows];
      	    }
        }
} 
/*******************************************************************/
rowsinputmapping()
{	int n1,n2,rows,cols,n; double s[512][512];

  	for(n1=0; n1 <= N; n1++) 
           for (n2=0; n2<N; n2++ ) 
               s[n1][n2] = x[n1][n2];

        for (cols=0; cols<N; cols++) {
  	  for(n=0; n < N/2; n++) {
     		 x[cols][n]     = s[cols][2*n];
     	 	x[cols][N-n-1] = s[cols][2*n+1];
         }
      }
	
 }  
/**********************************************************************/
ans_read()
{
	int rows,cols;
	double value;
	char file_name[50];
	FILE *fi,*fopen();


	printf("Enter name of file to be FCTransformed: ");
	scanf("%s",file_name);
     	if((fi=fopen(file_name,"r"))==NULL) {
		printf("cannot open file  '%s' \n",file_name);
		exit(1);
	}
        for (rows=0; rows<N; rows++) {
            for (cols=0; cols<N; cols++) {
                   if ((fscanf(fi,"%lf",&value)!=EOF))  {
		        x[cols][rows] = value;
                   }
             }
        } 
	fclose(fi);
}/* End of "ans_read" function */

/*-------------------------------------------------------------------*/
ans_write()
{
	int rows,cols;
	FILE *fo,*fopen();

	fo=fopen("DIT_FCT.OUT","w");
        for (rows=0; rows<N; rows++) {
            for (cols=0; cols<N; cols ++) {
              /* printf("%d  %d  %lf\n",cols,rows,x[cols][rows]);*/
                fprintf(fo,"%d  %d  %lf\n",cols,rows,x[cols][rows]);
            }
        }
 	fclose(fo);

	puts("  ");
	puts("Your transformed output file is:  DIT_FCT.OUT");
} /* end of "ans_write" function */
