/*Implementation of 2D Fast cosine transform (FCT) through the 1d algorithm of 
  Hou. Based in paper:

 H.S. Hou, "A fast recursive algorithm for computing the discrete cosine 
transform", IEEE Trans.  on Accoustics, Speech and Signal Processing, Vol. 35,
No. 10, pp. 1455-1461, Oct. 1987.

Written by:

C.A. Christopoulos
Vrije Universiteit Brussel
VUB - ETRO (IRIS)
Pleinlaan 2
1050 Brussels
Belgium
Tel:	++ 32 - 2 - 6412940
Fax:	++ 32 - 2 - 6412883
E-mail:	chchrist@etro.vub.ac.be
 
   18/10/1993 
*/


#include <math.h>
#include <stdio.h>
#define PI 3.141592653589

unsigned N,m,N0; double x[512][512],ct[4096];

main()
  {
		 double mtest = 1.;
		 int nl, nloops;

		printf("Enter number of points (N=power of 2): ");
		 scanf("%d",&N);
		 printf("Enter no of loops:");
		 scanf("%d",&nloops);
		 mtest=log(N+0.)/log(2.);
                 N0=N;
		 m=mtest;
		/* }*/

  ans_read();
   
      makecosinetable();
      two_d_throughHou();
  
 /* ans_write();*/
 }                      /* end of main */
/*************************************************************/
makecosinetable() /* 1D cosine table for N points */
  {
               int i,j,k,NN,mm,mN,p;
	       double xu, xd,PIT2=2*PI;
               p=0;

	       for (j=m; j>0; j--)
	       {
		    NN = 1<<j;
		    for (k=0; k<N; k = k+NN) {
			for (i=0; i<(NN>>1); i++)  {
			   ct[p++]= cos((PIT2/NN) * (i+0.25));
		 	 }
		    }
		}	      
  }/* end of cosine table*/

/*************************************************************/
 two_d_throughHou() /* 2d using the row-column approach */
  {
    int i,j,k,NN,mm,mN,rows,cols,p;
    register double xu, xd;
    double PIT2=2*PI;
 
   /* Transform rows */  
    rowsinputmapping();
    for (rows=0; rows<N; rows++) {
         p=0;
	 for (j=m; j>0; j--)
	       {
		    NN = 1<<j;
		    for (k=0; k<N; k = k+NN) {
			for (i=0; i<(NN>>1); i++)  {
			   mm = i+k;
			   mN=mm + (NN>>1);
			   xu = x[rows][mm] + x[rows][mN];
			   xd = x[rows][mm] - x[rows][mN];
			   xd = xd *  ct[p++];
			   x[rows][mm] = xu;
			   x[rows][mN] = xd;
			 }
		     }
		 }
     }

     for (j=2; j<m+1; j++)  {
	NN = 1<<j;
        for (k=0; k<(1<<m); k=k+NN)  {
		RLRrows((NN>>1)+k,j-1);
        }
     }
     rowsbitreversal(0,m);
      
 /* Transform columns */
  columnsinputmapping();	
  for (cols=0; cols<N; cols++) {
         p=0;
         for (j=m; j>0; j--)
	       {
		    NN = 1<<j;
		    for (k=0; k<N; k = k+NN) {
			for (i=0; i<(NN>>1); i++)  {
			   mm = i+k;
			   mN=mm + (NN>>1);
			   xu = x[mm][cols] + x[mN][cols];
			   xd = x[mm][cols] - x[mN][cols];
			   xd = xd *  ct[p++];
			   x[mm][cols] = xu;
			   x[mN][cols] = xd;
			 }
		     }
		 }
   }
   for (j=2; j<m+1; j++)  {
       NN = 1<<j;
       for (k=0; k<(1<<m); k=k+NN)  {
	   RLRcolumns((NN>>1)+k,j-1);
       }
   }
   columnsbitreversal(0,m);
    
}  /*  two_d_throughHou    */
/******************************************************************/
 RLRrows(val1,val2)
int val1,val2;
{
      float temp;
      int i,rows;

      rowsbitreversal(val1,val2);
      for (rows=0; rows<N; rows++) {
         temp = x[rows][val1];
      	for (i=val1+1; i<val1+(1<<val2); i++) {
	  temp = 2*x[rows][i] - temp;
	  x[rows][i] = temp;
        }
      }
     rowsbitreversal(val1,val2);
 }
/******************************************************************/
 rowsbitreversal(val1,val2) /* Bit reversal function */
  int val1,val2;
  {
   int v1, v2, v3,i,j,k,rows;
   double xt;
   double xx[512][512];

     for (rows=0; rows<N; rows++) 
	for(i=0; i<(1<<val2); i++) 
		 xx[rows][i] = x[rows][i+val1];

   for (rows=0; rows<N; rows++) {
   	v1 = (val2+1)>>1; v2 = 1 << v1; v3 = (1<<val2)-1-v2;
  	 j=0;
  	 for(i=1; i<=v3; i++){
	 	  k= ((1<<val2))>>1;
	  	  while(k<=j){
		  	 j=j-k; k=k>>1;
	   	  }
	   	j=j+k;
	 	  if(i<j){
		  	 xt=xx[rows][j];
		  	 xx[rows][j]=xx[rows][i];
		  	 xx[rows][i]=xt;
	   	  }
 	  }
    }
    for (rows=0; rows<N; rows++)
   	for (i=0; i<(1<<val2); i++)
    	  x[rows][i+val1] = xx[rows][i];

  } /*    End of 'bit_rev'    */
 /*****************************************************************/
RLRcolumns(val1,val2)
int val1,val2;
{
      float temp;
      int i,cols;

      columnsbitreversal(val1,val2);
      for (cols=0; cols<N; cols++) {
     	 temp = x[val1][cols];
     	 for (i=val1+1; i<val1+(1<<val2); i++) {
	 	temp = 2*x[i][cols] - temp;
	 	x[i][cols] = temp;
     	 }
      }
     columnsbitreversal(val1,val2);
 }

/*****************************************************************/
columnsbitreversal(val1,val2) /* Bit reversal function */
  int val1,val2;
  {
   int v1, v2, v3,i,j,k,cols;
   double xt;
   double xx[512][512];

   for (cols=0; cols<N; cols++) 
     for(i=0; i<(1<<val2); i++)
	 xx[i][cols] = x[i+val1][cols];
   
   for (cols=0; cols<N; cols++) {
 	  v1 = (val2+1)>>1; v2 = 1 << v1; v3 = (1<<val2)-1-v2;
  	 j=0;
  	 for(i=1; i<=v3; i++){
	 	  k= ((1<<val2))>>1;
	  	  while(k<=j){
		  	 j=j-k; k=k>>1;
	   	  }
	   	  j=j+k;
	   	  if(i<j){
			   xt=xx[j][cols];
		 	  xx[j][cols]=xx[i][cols];
		  	 xx[i][cols]=xt;
	  	  }
  	 }
    }
   for (cols=0; cols<N; cols++) 
  	 for (i=0; i<(1<<val2); i++)
      		x[i+val1][cols] = xx[i][cols];

  } /*    End of 'bit_rev'    */

/******************************************************************/
columnsinputmapping() /* 2d input mapping */
{
  
	int n1,n2,rows,cols,n; double s[512][512];

  	for(cols=0; cols<N; cols++) 
           for (n2=0; n2<N; n2++ ) 
               s[n2][cols] = x[n2][cols];

        for (cols=0; cols<N; cols++) {
  	    for(n=0; n < N/2; n++) {
     		 x[n][cols]     = s[2*n][cols];
     	 	x[N-n-1][cols] = s[2*n+1][cols];
      	    }
        }
} 
/*******************************************************************/
rowsinputmapping()
{	int n1,n2,rows,cols,n; double s[512][512];

  	for(rows=0; rows < N; rows++) 
           for (n2=0; n2<N; n2++ ) 
               s[rows][n2] = x[rows][n2];

        for (rows=0; rows<N; rows++) {
  	  for(n=0; n < N/2; n++) {
     		 x[rows][n]     = s[rows][2*n];
     	 	x[rows ][N-n-1] = s[rows][2*n+1];
         }
      }
	
 }  
/**********************************************************************/
ans_read()
{
	int rows,cols;
	double value;
	char file_name[50];
	FILE *fi,*fopen();


	printf("Enter name of file to be FCTransformed: ");
	scanf("%s",file_name);
     	if((fi=fopen(file_name,"r"))==NULL) {
		printf("cannot open file  '%s' \n",file_name);
		exit(1);
	}
        for (rows=0; rows<N; rows++) {
            for (cols=0; cols<N; cols++) {
                   if ((fscanf(fi,"%lf",&value)!=EOF))  {
		        x[cols][rows] = value;
                   }
             }
        } 
	fclose(fi);
}/* End of "ans_read" function */

/*-------------------------------------------------------------------*/
ans_write()
{
	int rows,cols;
	FILE *fo,*fopen();

	fo=fopen("DIT_FCT.OUT","w");
        for (rows=0; rows<N; rows++) {
            for (cols=0; cols<N; cols ++) {
               /*printf("%d  %d  %lf\n",cols,rows,x[cols][rows]);*/
                /*fprintf(fo,"%d  %d  %lf\n",cols,rows,x[cols][rows]);*/
            }
        }
 	fclose(fo);

	puts("  ");
	puts("Your transformed output file is:  DIT_FCT.OUT");
} /* end of "ans_write" function */
