/* Computation of the two-dimensional FCT through the 1D decimation in 
time (or decimation in space (DIS)) algorithm (no pruning is implemented). 
   The computation is done through the input-reordering algorithm. Based on paper:
 A.N. Skodras, "Fast discrete cosine transform pruning", IEEE Trans. on 
Signal Processing. Accepted for pulication (I think it will appear on
June or July 1994).

Written by:

C.A. Christopoulos
Vrije Universiteit Brussel
VUB - ETRO (IRIS)
Pleinlaan 2
1050 Brussels
Belgium
Tel:	++ 32 - 2 - 6412940
Fax:	++ 32 - 2 - 6412883
E-mail:	chchrist@etro.vub.ac.be
 
   20/10/1993 
*/
#include <math.h>
#include <stdio.h>
#define PIO2 1.570796327

unsigned int N,N0,m,r[512][512],MASK[2],tst,nloops,loop;
double x[512][512], ct[1024];

main()
{
 double mtest = 1.;

 while (mtest - m != 0)  {
   printf("Enter number of points (N = power of 2): ");
   scanf("%d",&N );
   mtest = log(N+0.) / log(2.);
   m = mtest;
   }

printf("Enter number of loops:");
scanf("%d",&nloops);
N0=N;
MASK[0] = 0;
MASK[1] = ~((-1) << m);

lut_of_cos();

ans_read();
     p_fct();
 

ans_write();

}  /* end of 'main'  */


/*------------------------------------------------------------------*/
void bitreversalcolumns()
{
unsigned int i,j,l,rows;

for (rows=0; rows<N; rows++) {
   l=1; 
   r[0][rows]=0;
   for(i=1; i < m; i++)
   {
       for(j=0; j < l; j++) {r[j][rows]<<= 1; r[j+l][rows]=r[j][rows] + 1; }
       l <<= 1;
   }
} /* end for rows */
}

/*************************************************************/
void bitreversalrows()
{
unsigned int i,j,l,cols;


for (cols=0; cols<N; cols++) {
   l=1; 
   r[cols][0]=0;
   for(i=1; i < m; i++)
   {
       for(j=0; j < l; j++) {r[cols][j]<<= 1;r[cols][j+l]=r[cols][j] +  1;  }
       l <<= 1;
   }
} /* end for cols */

}/* End of bit-reversal */
/*--------------------------------------------------------------*/
firo3columns()
{
unsigned int i,j,eo,group,nog,p,q,F,M,a,b,rows,cols;
double t;

M=m;

bitreversalcolumns();

for (rows=0; rows<N; rows++) {
        M=m;
	eo = M%2; M = m>>1;
	group = nog = 1<<M;
	if(eo==1) M++;

/*..................... M=even/odd ..........................*/

	for(i=0; i<(nog-1); i++)
	 {
  		 F=0; q=i<<M; p=q>>1;
  		 for(j=1; j<group; j++)
   	 	{
    	 		 F=1-F; q++;
                         a=((r[p][rows]<<1)^MASK[F]); /* CC*/
                         b=q;  /*CC*/
                         t=x[a][rows]; /*CC*/
        		 x[a][rows] = x[b][rows]; /*CC*/
                         x[b][rows] = t;/*CC*/
     			/* swap( ((r[p][rows]<<1)^MASK[F]), q );*/
      			p += F;
   		 }
 		  group--;
	 }

	if(eo==0) ; else

/*....................... M=odd ..........................*/

	{ group=nog;

 	 for(i=1; i<nog; i++)
  	 {
   		 F=0; q=i<<M; p=q>>1; p--; q--;
     		for(j=1; j<group; j++)
      		{
			q--;
                         a=((r[p][rows]<<1)^MASK[F]); /* CC*/
                         b=q;  /*CC*/
                         t=x[a][rows]; /*CC*/
        		 x[a][rows] = x[b][rows]; /*CC*/
                         x[b][rows] = t;/*CC*/

			/*swap( ((r[p][rows]<<1)^MASK[F]), q );*/
			F=1-F;   p -= F;
     		 }
    		 group--;
  	 }
	} /* end of 'if' statement */
    } /* end for rows */
}/* end */
/***********************************************************/
firo3rows()
{
   unsigned int i,j,eo,group,nog,p,q,F,M,a,b,rows,cols;
   double t;

M=m;
bitreversalrows();

/* Input reordering for the rows */
for (cols=0; cols<N; cols++) {
        M=m;
	eo = M%2; M = m>>1;
	group = nog = 1<<M;
	if(eo==1) M++;

/*..................... M=even/odd ..........................*/

	for(i=0; i<(nog-1); i++)
	 {
  		 F=0; q=i<<M; p=q>>1;
  		 for(j=1; j<group; j++)
   	 	{
    	 		 F=1-F; q++;
                         a=((r[cols][p]<<1)^MASK[F]); /* CC*/
                         b=q;  /*CC*/
                         t=x[cols][a]; /*CC*/
        		 x[cols][a] = x[cols][b]; /*CC*/
                         x[cols][b] = t;/*CC*/
                        /* swap( ((r[cols][p]<<1)^MASK[F]), q );*/
      			p += F;
   		 }
 		  group--;
	 }

	if(eo==0) ; else

/*....................... M=odd ..........................*/

	{ group=nog;

 	 for(i=1; i<nog; i++)
  	 {
   		 F=0; q=i<<M; p=q>>1; p--; q--;
     		for(j=1; j<group; j++)
      		{
			q--;
                         a=((r[cols][p]<<1)^MASK[F]); /* CC*/
                         b=q;  /*CC*/
                         t=x[cols][a]; /*CC*/
        		 x[cols][a] = x[cols][b]; /*CC*/
                         x[cols][b] = t;/*CC*/
			F=1-F;   p -= F;
     		 }
    		 group--;
  	 }
	} /* end of 'if' statement */
   } /* end for rows */

} /* End of input reordering algorithm */

 /********************************************************************/
lut_of_cos()
{
unsigned int e,i,k,l,p,t,inc,len,mm1,et[1024];

p=0; mm1=m-1; e=1;

for(k=0; k<m; k++)
   {
   len=1; inc=N; i=0;
   et[i]=e; i++; ct[p]=2 * cos(PIO2 * e / N); p++;
   for(t=0; t<mm1; t++)
      {
      for(l=0; l<len; l++)
	 {
	 et[i] = et[l] + inc;
	 ct[p] = 2 * cos(et[i] * PIO2 /N);
	 i++;  p++;
	 }
      len = len << 1; inc=inc >> 1;
      }
   e = e << 1; mm1=mm1-1;
   }

} /* End of 'lut_of_cos'function */

/*------------------------------------------------------------------*/
post_addscolumns()
{
     int step,loops,k,ep,j,i,l,cols,rows;

  for (cols=0; cols<N; cols++) 
         for (rows=1; rows<N; rows++) 
             x[rows][cols] = x[rows][cols]/2;

      for (cols=0; cols<N; cols++) {   
        step =N;
	loops = 1;
	for (k=1; k<m; k++)  {
	     step = step>>1;
	     ep = step>>1;
	     loops = loops <<1;
	     for (j=0; j<(step>>1); j++) {
	        l=ep;
		for (i=1; i<loops; i++)  {
                      x[l+step][cols] = x[l+step][cols] - x[l][cols];
                    l =l+step;
               
	         }
		 ep +=1;
	     }
	}
    }
} /* end of post_addsrows */
/*******************************************************************/
 post_addsrows()
{
     int step,loops,k,ep,j,i,l,cols,rows;

     for (rows=0; rows<N; rows++) 
         for (cols=1; cols<N; cols++) 
             x[rows][cols] = x[rows][cols]/2;

     for (rows=0; rows<N;rows++) {
       step =N;
	 loops = 1;
	   for (k=1; k<m; k++)  {
                step = step>>1;
	        ep = step>>1;
            	loops = loops <<1;
		for (j=0; j<(step>>1); j++) {
	            l=ep;
		    for (i=1; i<loops; i++)  {  
			   x[rows][l+step] = x[rows][l+step]-x[rows][l];
			    l =l+step;
	            }
		    ep +=1;
		  }
	     }
        }
     
} /* end of post additions */

/*------------------------------------------------------------------*/


p_fct()
{
unsigned int i,j,k,l,p,q,r,Bs, bB,bls,ble,ncb,nib,rows,cols,tst;
double a,c;

firo3rows(); /* inputreordering for the rows */
for (rows=0; rows<N; rows++) {
   Bs = N;  bB = bls = 1;  p = 0;

  for(k=0; k<m; k++) {
        Bs=Bs>>1; ble=0; bls=bls<<1;

        for(j=0; j<Bs; j++)
        {
            l=ble; c=ct[p]; p++;
            for(i=0; i<bB; i++)
	   {
		 a=x[rows][l];
		 x[rows][l] += x[rows][l+bB];
	 	 x[rows][l+bB]=(a-x[rows][l+bB]) * c;
		 l++;
	   }

           ble += bls;
        }
 	bB=bB<<1;
  }
 } /* end rows*/
  post_addsrows();

/*** Do the columns ***/

firo3columns();
for (cols=0; cols<N; cols++) {
  Bs = N;  bB = bls = 1;  p = 0;
    for(k=0; k<m; k++)
   {
   Bs=Bs>>1; ble=0; bls=bls<<1;

   for(j=0; j<Bs; j++)
      {
      l=ble; c=ct[p]; p++;
      for(i=0; i<bB; i++)
	 {
	 a=x[l][cols];
	 x[l][cols] += x[l+bB][cols];
	 x[l+bB][cols]=(a-x[l+bB][cols]) * c;
	 l++;
	 }

         ble += bls;
     }
  bB=bB<<1;
  }
 } /* end columns*/
post_addscolumns();


} /* End of 'p_fct' function */

/*------------------------------------------------------------------*/
ans_read()
{
	int rows,cols;
	double value;
	char file_name[50];
	FILE *fi,*fopen();


	printf("Enter name of file to be FCTransformed: ");
	scanf("%s",file_name);
     	if((fi=fopen(file_name,"r"))==NULL) {
		printf("cannot open file  '%s' \n",file_name);
		exit(1);
	}
        for (rows=0; rows<N; rows++) {
             for (cols=0; cols<N; cols++) {
                   if ((fscanf(fi,"%lf",&value)!=EOF))  {
		        x[cols][rows] = value;
                   }
             }
        } 
	fclose(fi);
}/* End of "ans_read" function */

/*-------------------------------------------------------------------*/
ans_write()
{
	int rows,cols;
	FILE *fo,*fopen();

	fo=fopen("DIT_2D_FCT.OUT","w");
        for (rows=0; rows<N; rows++) {
            for (cols=0; cols<N; cols ++) {
                /* printf("%d  %d  %lf\n",cols,rows,x[cols][rows]);*/
                 fprintf(fo,"%d  %d  %lf\n",cols,rows,x[cols][rows]);
            }
        }
 	fclose(fo);

	puts("  ");
	puts("Your transformed output file is:  DIT_2D_FCT.OUT");
} /* end of "ans_write" function */
