/*Implementation of 2D FCT through the 1D decimation in 
frequency (DIF) split radix algorithm from paper:

 C.A. Christopoulos and A. N. Skodras, "On the computation of the fast cosine
transform", Proceedings of the European Conference on Circuit Theory and Design
(ECCTD'93), Davos, Switzerland, pp. 1037-1042, Aug.30-Sept. 3, 1993.
 
 Written by C. A. Christopoulos
Vrije Universiteit Brussel
VUB - ETRO (IRIS)
Pleinlaan 2
1050 Brussels
Belgium
Tel:	++ 32 - 2 - 6412940
Fax:	++ 32 - 2 - 6412883
E-mail:	chchrist@etro.vub.ac.be


   16 /10/1993
*/


#include <math.h>
#include <stdio.h>
#define PI 3.141592653589

unsigned N,m,N0; double x[512][512],ct[512];

main()
  {
		 double mtest = 1.;
		 int nl, nloops;

		printf("Enter number of points (N=power of 2): ");
		 scanf("%d",&N);
		 printf("Enter no of loops:");
		 scanf("%d",&nloops);
		 mtest=log(N+0.)/log(2.);
                 N0=N;
		 m=mtest;
		/* }*/

  ans_read();
     makecosinetable();
      two_d_throughsrfct();
  
/*  ans_write();*/
 }                      /* end of main */
/*************************************************************/

makecosinetable() /* 1D cosine table for N points */
{
   int n2,n4,k,j,p;
   double e;

	n2 = N << 1;
        p=0;

	for (k = 1; k < m; k++) {

		e  = PI / n2;
		n2 = n2 >> 1;
		n4 = n2 >> 2;

		for (j = 0; j < n4; j++) {

			ct[p++] = cos(((j<<2)+1)*e);
			ct[p++] = sin(((j<<2)+1)*e);
			ct[p++] = cos((((j<<2)+1)<<1)*e);

		}
	}

	ct[p++] = cos(PI/4.0);


}/* end of cosine table*/

/*************************************************************/
two_d_throughsrfct() /* 2d using the row-column approach */
  {
  	int n2,n4,k,is,j,id,i0,i1,i2,i3,rows,cols,p;
	double a,e;
	register double t1,t2;
   
  rowsinputmapping();
  for (rows=0; rows<N; rows++) {
        p=0;	
 	n2 = N << 1;
	for (k = 1; k < m; k++) {
		e  = PI / n2;
		n2 = n2 >> 1;
		n4 = n2 >> 2;

		for (j = 0; j < n4; j++) {

			is = j;
			id = n2 << 1;

again1:   
			for (i0 = is; i0 < N; i0 +=id) {

				i1 = i0 + n4;
				i2 = i1 + n4;
				i3 = i2 + n4;

				t1 = 2 * (x[rows][i0] - x[rows][i2]) *ct[p] ;
				t2 = (-2) * (x[rows][i1] - x[rows][i3]) * ct[p+1];

				x[rows][i0] = x[rows][i0] + x[rows][i2];
				x[rows][i1] = x[rows][i1] + x[rows][i3];
				x[rows][i2] = t1 + t2;
				x[rows][i3] = 2 * (t1 - t2) * ct[p+2];

				/* if at the m-1 stage, divide by 2 */
				if (k==m-1) {
					x[rows][i2] = x[rows][i2]/2;
					x[rows][i3] = x[rows][i3]/2;
				}
			}

			is = (id << 1) - n2 + j;
			id = id << 2;
			if ( is < N ) goto again1;
                        p+=3;
		}
	}
	
	/* last stage */
	is = 0;
	id = 4;
l_again1: 
	for( i0 = is; i0 < N; i0 += id) {
		i1 = i0 + 1;
		t1 = x[rows][i0];
		x[rows][i0] = t1 + x[rows][i1];
		/* divide the upper element by 2 */
		if (i0 != 0) x[rows][i0] = x[rows][i0]/2;
		x[rows][i1] = (t1 - x[rows][i1]) * ct[p];
	}

	is = (id << 1) - 2;
	id =  id << 2;
	if( is < N ) goto l_again1;

    }/* end of for rows */
    rowsbitreversal();
    rowspostadditions();

  columnsinputmapping();	
  for (cols=0; cols<N; cols++) {
         p=0;
	 p=0;	
 	n2 = N << 1;
	for (k = 1; k < m; k++) {
		e  = PI / n2;
		n2 = n2 >> 1;
		n4 = n2 >> 2;

		for (j = 0; j < n4; j++) {

			is = j;
			id = n2 << 1;

again2:   
			for (i0 = is; i0 < N; i0 +=id) {

				i1 = i0 + n4;
				i2 = i1 + n4;
				i3 = i2 + n4;

				t1 = 2 * (x[i0][cols] - x[i2][cols]) *ct[p] ;
				t2 = (-2) * (x[i1][cols]-x[i3][cols])* ct[p+1];

				x[i0][cols] = x[i0][cols] + x[i2][cols];
				x[i1][cols] = x[i1][cols] + x[i3][cols];
				x[i2][cols] = t1 + t2;
				x[i3][cols] = 2 * (t1 - t2) * ct[p+2];

				/* if at the m-1 stage, divide by 2 */
				if (k==m-1) {
					x[i2][cols] = x[i2][cols]/2;
					x[i3][cols] = x[i3][cols]/2;
				}
			}

			is = (id << 1) - n2 + j;
			id = id << 2;
			if ( is < N ) goto again2;
                        p+=3;
		}
	}
	
	/* last stage */
	is = 0;
	id = 4;
l_again2: 
	for( i0 = is; i0 < N; i0 += id) {
		i1 = i0 + 1;

		t1 = x[i0][cols];
		x[i0][cols] = t1 + x[i1][cols];
		/* divide the upper element by 2 */
		if (i0 != 0) x[i0][cols] = x[i0][cols]/2;
		x[i1][cols] = (t1 - x[i1][cols]) * ct[p];
	}

	is = (id << 1) - 2;
	id =  id << 2;
	if( is < N ) goto l_again2;
      
  }/* end of for cols */
    columnsbitreversal();
    columnspostadditions();


  }  /*    end of SR_FCT    */
/******************************************************************/
columnspostadditions()
{
     int step,loops,k,ep,j,i,l,cols,rows;

    for (cols=0; cols<N; cols++) {
        step =N;
	loops = 1;
	for (k=1; k<m; k++)  {
            step = step>>1;
	     ep = step>>1;
	     loops = loops <<1;
	     for (j=0; j<(step>>1); j++) {
	        l=ep;
                for (i=1; i<loops; i++)  {
                     x[l+step][cols] = x[l+step][cols] - x[l][cols];
                   l =l+step;
                }
		 ep +=1;
	     }
	}
    }
 } /* end of rowspostadditions */   
/*****************************************************************/
rowspostadditions()
{
 int step,loops,k,ep,j,i,l,cols,rows;


      /* Postaditions for the columns */
      for (rows=0; rows<N; rows++) {
         step =N;
	 loops = 1;
	   for (k=1; k<m; k++)  {
                step = step>>1;
	        ep = step>>1;
            	loops = loops <<1;
		for (j=0; j<(step>>1); j++) {
	            l=ep;
                    for (i=1; i<loops;i++)  {
                       x[rows][l+step] = x[rows][l+step]-x[rows][l];
                       l =l+step;
	            }
		    ep +=1;
		  }
	     }
       }
        
     
} /* end of columnspostadditions */

/******************************************************************/
rowsbitreversal() /* 2d bit reversal */
{
    int v1, v2, v3,i,j,k,cols,rows;
    double xt;

    /* revesre rows */
     for (cols =0; cols<N; cols ++) {   
	  v1 = (m+1)/2; 
	  v2 = 1 << v1; 
	  v3 = N-1-v2;
	  j=0;
	  for(i=1; i<=v3; i++){
	    	k= N>>1;
		while(k<=j){
			j=j-k; 
			k=k>>1;
		}
		j +=k;
		if(i<j){
			xt=x[cols][j];
			x[cols][j]=x[cols][i];
			x[cols][i]=xt;
		}
	  }
     }
}
/*****************************************************************/
columnsbitreversal()
 {   
  int v1, v2, v3,i,j,k,cols,rows;
    double xt;
    /* reverse columns */
     for (rows =0; rows<N; rows ++) {   
	  v1 = (m+1)/2; 
	  v2 = 1 << v1; 
	  v3 = N-1-v2;
	  j=0;
	  for(i=1; i<=v3; i++){
	    	k= N>>1;
		while(k<=j){
			j=j-k; 
			k=k>>1;
		}
		j +=k;
		if(i<j){
			xt=x[j][rows];
			x[j][rows]=x[i][rows];
			x[i][rows]=xt;
		}
	  }
     }

} 
/******************************************************************/
columnsinputmapping() /* 2d input mapping */
{
  
	int n1,n2,rows,cols,n; double s[512][512];

  	for(n1=0; n1 <= N; n1++) 
           for (n2=0; n2<N; n2++ ) 
               s[n1][n2] = x[n1][n2];

        for (rows=0; rows<N; rows++) {
  	    for(n=0; n < N/2; n++) {
     		 x[n][rows]     = s[2*n][rows];
     	 	x[N-n-1][rows] = s[2*n+1][rows];
      	    }
        }
} 
/*******************************************************************/
rowsinputmapping()
{	int n1,n2,rows,cols,n; double s[512][512];

  	for(n1=0; n1 <= N; n1++) 
           for (n2=0; n2<N; n2++ ) 
               s[n1][n2] = x[n1][n2];

        for (cols=0; cols<N; cols++) {
  	  for(n=0; n < N/2; n++) {
     		 x[cols][n]     = s[cols][2*n];
     	 	x[cols][N-n-1] = s[cols][2*n+1];
         }
      }
	
 }  
/**********************************************************************/
ans_read()
{
	int rows,cols;
	double value;
	char file_name[50];
	FILE *fi,*fopen();


	printf("Enter name of file to be FCTransformed: ");
	scanf("%s",file_name);
     	if((fi=fopen(file_name,"r"))==NULL) {
		printf("cannot open file  '%s' \n",file_name);
		exit(1);
	}
        for (rows=0; rows<N; rows++) {
            for (cols=0; cols<N; cols++) {
                   if ((fscanf(fi,"%lf",&value)!=EOF))  {
		        x[cols][rows] = value;
                   }
             }
        } 
	fclose(fi);
}/* End of "ans_read" function */

/*-------------------------------------------------------------------*/
ans_write()
{
	int rows,cols;
	FILE *fo,*fopen();

	fo=fopen("DIT_FCT.OUT","w");
        for (rows=0; rows<N; rows++) {
            for (cols=0; cols<N; cols ++) {
              /* printf("%d  %d  %lf\n",cols,rows,x[cols][rows]);*/
                fprintf(fo,"%d  %d  %lf\n",cols,rows,x[cols][rows]);
            }
        }
 	fclose(fo);

	puts("  ");
	puts("Your transformed output file is:  DIT_FCT.OUT");
} /* end of "ans_write" function */
