 /*
	RECURSIVE PRUNING OF THE TWO-DIMENSIONAL DCT 

VECTOR - RADIX DECIMATION IN SPACE FAST COSINE TRANSFORM -RECURSIVE PRUNING

  Real two dimensional decimation in space fast cosine transform
	pruning algorithms.

Based on paper "Recursive pruning of 2D DFT with 3D
  signal processin applications", IEEE Trans. on Signal Processsing, 
  March 1993.

  The program is extension of the vector-radix DIT pruning.  N0 now is the 
   number of points to be computed in
  the zig-zag scanning order (N0=size of triangular). 
  If N0=2N, all points are computed. If  
  N0>N, the zig-zag is continued at the lower half of the NxN data.

  The post-additions stage is not good, because it does post additions 
  even in non-required points, i.e it does the post additions in the 
  NxN data.

Based on paper:
C.A. Christopoulos and A.N. Skodras, "Pruning the two-dimensional fast cosine 
transform", To be presented at the VII European Signal Processing Conference ,
September 13-16, 1994, Edinburgh, Scotland.

  
Written by:

 Charilaos Christopoulos
Vrije Universiteit Brussel
  VUB - ETRO (IRIS)
  Pleinlaan 2
  1050 Brussels
  Belgium

 Tel:++32-2-641 2940
 Fax:++32-2-641 2883
 Email:chchrist@etro.vub.ac.be


   23/10/1993

*/


#include <math.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#define TRUE 1
#define FALSE 0
#define DESIRED 1
#define level1  1   /*0x1   binary 0001 */
#define level2  8   /*0x8  binary 1000 */
#define level3  2   /*0x2  binary 0010 */
#define level4  4   /*0x4   binary 0100 */
 
#define PI 3.141592653589
#define PIO2 1.5707966327
unsigned int nBy2,marker,maskIndex,prunecounter;
unsigned int quadrantMask[4],doAllButterflies=15;
unsigned int FinalOutputMatrix[256][256]={0};
unsigned int OutputPruneMatrix[8][256][256];
unsigned int OutputPruneMatrix1d[8][65536];

int PassNumber;
int N,m,N0, count=0; 
int point,counter;
double x[512][512];
unsigned int r[256][256];
double cosine_array[4096]; 
unsigned int MASK[2],tst;
double ct[256]; /* cosine table for the 1D case*/
double ct2d[500000]; /* cosine table to be used for the 2D case */
/***************************************************************/
main()
{
        int k,loop,noloops,zigzag,i,j;
	double mtest = 1;
        printf("Enter dimension (N=power of 2): ");
	scanf("%d", &N );
	printf("Required points at each dimension?");
        scanf("%d",&N0); /* N0 must be power of 2 */
 
                mtest=log(N+0.)/log(2.);
		m=mtest;
       
       MASK[0] = 0;
       MASK[1] = ~((-1) << m);
        quadrantMask[0] = level1;
        quadrantMask[1] = level2;
        quadrantMask[2] = level3;
        quadrantMask[3] = level4;
        doAllButterflies = 15;
        zigzag=N0;
	for (i=0; i< N0; i++) {
             for (j=0; j<zigzag; j++) {
                FinalOutputMatrix[i][j] =1;
             }
             zigzag --;
        }
	SetOutputPruneMatrixRecursive(FinalOutputMatrix,N,OutputPruneMatrix);

        ans_read();
        lut_of_cos();
	expand1d_lookup_table();
        make2Darray();	

	vector_radix_FCT();
 
         ans_write();
}                      /* end of main */
/*******************************************************************/
SetOutputPruneMatrixRecursive(L2Matrix,NoPoints,OutputPruneMatrix)
    unsigned int L2Matrix[256][256];
    unsigned int NoPoints;
    unsigned int OutputPruneMatrix[8][256][256];

{
     unsigned int k1,k2,i,j,maskIndex,allButterflies,passNumber;
     unsigned int counter;
    counter =0;

    passNumber= log ((float)NoPoints)/log(2.);;
    if (passNumber==1) { 
          OutputPruneMatrix[1][0][0] = doAllButterflies;
          OutputPruneMatrix1d[1][0] = doAllButterflies;
    }
    else
    {
         nBy2 = NoPoints/2;
         allButterflies=TRUE;
    
         for (k1=0; k1<nBy2; k1++) {
              for (k2=0; k2<nBy2; k2++) {
                 /* 
                   Examine each output point in the radix-(2x2)
                   butterfly. If the point is required, add it to 
                   the marker.
                 */
                  maskIndex = 0;
                  marker =0;
                  for (i=k1; i<NoPoints; i+=nBy2) {
                         for (j=k2; j<NoPoints; j+=nBy2) {
                             if (L2Matrix[i][j] != 0) {
                                   marker = marker|quadrantMask[maskIndex];
                              }
                              maskIndex ++;
                         }
                   }
                   /* 
                      The marker now indicates which of the four radix-(2x2)
                      output points are required. Assign the marker to the
                      appropriate location in the OutputPruneMatrix.
                   */
                    OutputPruneMatrix[passNumber][k1][k2] = marker;
                    /*OutputPruneMatrix1d[passNumber][counter] = marker;*/
                   
                    counter++;
/*printf("passNumber = %d  marker = %d    %d \n",passNumber,marker,OutputPruneMatrix[passNumber][k1][k2]); */
                    if ( (allButterflies == TRUE) && ((marker!=15)) )
                                 allButterflies = FALSE;
               } /* for k2 */
          } /* for k1 */
          if (allButterflies == TRUE) {
                    OutputPruneMatrix[PassNumber][0][0] = doAllButterflies;
                   /* OutputPruneMatrix1d[passNumber][0] = doAllButterflies;*/
          }
          /* 
             Now find the OutoutPruneMatrix for the previous FCT pass 
             using the OutputPruneMatrix of this pass 
          */
          prunecounter =0;
              
          SetOutputPruneMatrixRecursive(OutputPruneMatrix[passNumber],NoPoints/2,
                OutputPruneMatrix);
       
      } /*else */

    
  }/* end of procedure 'SetOutputPruneMatrixRecursive' */
 /*******************************************************************/
vector_radix_FCT()
{
 int stage;
 unsigned int marker1, marker2, marker3, marker4, pointer;
 int N1,M1,N2, M2;
 int i,j,ii,jj,nr,nc,nr1,nc1,sign,intme;
   int NN,Ni,k1,k2,k3,k33,yi,yi1,yi2,xj,xj1,xj2,mmax,istep,step;
 double S0,S1,S2,S3;
 float sum1,sum2,diff1,diff2;

	NN = N/2;
           firo3();
 	
/* decimation in time DCT */

/* Perform the first stage of the transform */
istep=2;
step=0;
 
for (yi=0; yi<N; yi+=2) {
        yi1 = yi;
   	yi2 = yi1+1;
  	for (xj=0; xj<N; xj +=2) {
 		xj1=xj;
      		xj2=xj1+1;
 
    		S0=x[yi1][xj1]; 
		S1=x[yi2][xj1];
		S2=x[yi1][xj2];
     	 	S3=x[yi2][xj2];  
 
      		sum1 = S0 + S1;
      		sum2 = S2 + S3;
      		diff1= S0 - S1;
      		diff2= S2 - S3;
 
      		x[yi1][xj1] = sum1+sum2;
      		x[yi2][xj1] = (diff1+diff2)*ct2d[step++];
      		x[yi1][xj2] = (sum1-sum2)* ct2d[step++];
      		x[yi2][xj2] = (diff1-diff2)*ct2d[step++] ;
	}
}
 
/* Perfrom the remaining stages of the transform */
stage=0;
mmax=2;
 while (N>mmax) {
        stage++;
         istep=2*mmax;
	for (k1=0; k1<mmax; k1++) {
		for (k2=0; k2<mmax; k2++) {
		    for (yi=k1; yi<N; yi+=istep) {
			yi1=yi; 
			yi2=yi1+mmax;
			for (xj=k2; xj<N; xj+=istep) {
				pointer =0;
				xj1=xj;
				xj2=xj1+mmax;
		
				S0=x[yi1][xj1];
				S1=x[yi2][xj1];
				S2=x[yi1][xj2];
				 S3=x[yi2][xj2];

				sum1=S0+S1;
				sum2 = S2+S3;
				diff1=S0-S1;
				diff2 = S2-S3;

		  		 marker1 = OutputPruneMatrix[stage+1][k1][k2] & level1;
      		  		 marker2 = OutputPruneMatrix[stage+1][k1][k2] & level2;
		  		 marker3 = OutputPruneMatrix[stage+1][k1][k2] & level3;
      		  		 marker4 = OutputPruneMatrix[stage+1][k1][k2] & level4;
		
		   		 if (marker1 == level1)
					 x[yi1][xj1] = sum1+sum2;
		    		if (marker3==level3) 
					  x[yi2][xj1]=(diff1+diff2)*ct2d[step];
		    		if (marker2==level2)
			 		 x[yi1][xj2]=(sum1-sum2)*ct2d[step+1] ;
		    		if (marker4==level4)
			 		 x[yi2][xj2]=(diff1-diff2)*ct2d[step+2] ;
			        step +=3; 
			  }
		      }  
		}  
	}  
	mmax = istep;
}  

 
post_adds();
printf("\nEND ");

}
 
/*******************************************************************/
make2Darray()
{
 int ND1,MD1;
 int N1,M1,N2, M2;
 int i,j,ii,jj,nr,nc,nr1,nc1,sign,intme;
 int NN,Ni,k1,k2,k3,k33,yi,yi1,yi2,xj,xj1,xj2,mmax,istep,cos_step,step;
 
	NN = N/2;
 ND1=0;
 MD1=0;
 cos_step=0;
 step=0;
 
for (yi=0; yi<N; yi+=2) {
        yi1 = yi+MD1;
   	yi2 = yi1+1;
  	for (xj=0; xj<N; xj +=2) {
        	xj1=xj+ND1;
      		xj2=xj1+1;
 	 	ct2d[step ++] =  cosine_array[yi2]*cosine_array[xj1];
      		ct2d[step ++] =  cosine_array[yi1]*cosine_array[xj2];
      		ct2d[step ++] =  cosine_array[yi2]*cosine_array[xj2];
	}
}
/* Find cosines for the remaining stages of the transform */
mmax=2;
 while (N>mmax) {
       cos_step +=N;
       istep=2*mmax;
	Ni=N/istep;

	for (k1=0; k1<mmax; k1++) {
	    for (k2=0; k2<mmax; k2++) {
		  for (yi=k1; yi<N; yi+=istep) {
			yi1=yi+MD1; 
			yi2=yi1+mmax;
			for (xj=k2; xj<N; xj+=istep) {
				xj1=xj+ND1;
				xj2=xj1+mmax;
		 			
ct2d[step ++]=cosine_array[cos_step+yi2]*cosine_array[cos_step+xj1];
ct2d[step ++]=cosine_array[cos_step+yi1]*cosine_array[cos_step+xj2];
ct2d[step ++]=cosine_array[cos_step+yi2]*cosine_array[cos_step+xj2];

				} /* x*/
			} /*y */
		} /* k2 */
	} /* k1 */
	mmax = istep;
} /* while */
  
}

/***************************************************************************/
 lut_of_cos() /*generate look-up table of cosines for the 1D-case */
{            /* as in the DIT FCT */
unsigned int e,i,k,l,p,t,inc,len,mm1,et[256];

p=0; mm1=m-1; e=1;

for(k=0; k<m; k++)
   {
   len=1; inc=N; i=0;
   et[i]=e; i++; ct[p]=2 * cos(PIO2 * e / N); p++;
   for(t=0; t<mm1; t++)
      {
      for(l=0; l<len; l++)
	 {
	 et[i] = et[l] + inc;
	 ct[p] = 2 * cos(et[i] * PIO2 /N);
	 i++;  p++;
	 }
      len = len << 1; inc=inc >> 1;
      }
   e = e << 1; mm1=mm1-1;
   }
} /* End of 'lut_of_cos'function */
 /******************************************************************/
expand1d_lookup_table() /* The lookup table generated above is expanded */
{                       /* the array is now of length m*N; */
unsigned int i,j,k,l,p,q,r,Bs, bB,bls,ble,ncb,nib,value,step;
double a,c;
value=0;

Bs = N;  bB = bls = 1;  p = 0; step=0;

 for(k=0; k<m; k++)
   {
   Bs=Bs>>1; q=N/bB; r=N%bB; ble=step; bls=bls<<1;

   if(q==1) {ncb=r; nib=bB-r;}
   if(q<1)  {ncb=0; nib=N;  } 
   if(q>1)  {ncb=bB; nib=0;  }

   for(j=0; j<Bs; j++)
      {
      l=ble; c=ct[p]; p++;
      for(i=0; i<ncb; i++)
         {
         cosine_array[l+step] = 1.0;
         cosine_array[step+l+bB]=c;
         value++;
         l++;
         }

         ble += bls;
     }
     bB=bB<<1;
     step += N/2;
  }
} /* expand1d_lookup_table */
/************************************************************/

 bitreversalrows()
{
unsigned int i,j,l,rows;

for (rows=0; rows<N; rows++) {
   l=1; 
   r[0][rows]=0;
   for(i=1; i < m; i++)
   {
       for(j=0; j < l; j++) {r[j][rows]<<= 1; r[j+l][rows]=r[j][rows] + 1; }
       l <<= 1;
   }
} /* end for rows */
}
/*************************************************************/
bitreversalcolumns()
{
unsigned int i,j,l,cols;
 for (cols=0; cols<N; cols++) {
   l=1; 
   r[cols][0]=0;
   for(i=1; i < m; i++)
   {
       for(j=0; j < l; j++) {r[cols][j]<<= 1;r[cols][j+l]=r[cols][j] +  1;}
       l <<= 1;
   }
} /* end for cols */
}/* End of bit-reversal */
/*--------------------------------------------------------------*/
firo3()
{
unsigned int i,j,eo,group,nog,p,q,F,M,a,b,rows,cols;
double t;

 M=m;
 bitreversalrows();
 
for (rows=0; rows<N; rows++) {
        M=m;
	eo = M%2; M = m>>1;
	group = nog = 1<<M;
	if(eo==1) M++;

/*..................... M=even/odd ..........................*/

	for(i=0; i<(nog-1); i++)
	 {
  		 F=0; q=i<<M; p=q>>1;
  		 for(j=1; j<group; j++)
   	 	{
    	 		 F=1-F; q++;
                         a=((r[p][rows]<<1)^MASK[F]); /* CC*/
                         b=q;  /*CC*/
                         t=x[a][rows]; /*CC*/
        		 x[a][rows] = x[b][rows]; /*CC*/
                         x[b][rows] = t;/*CC*/
     		p += F;
   		 }
 		  group--;
	 }

	if(eo==0) ; else

/*....................... M=odd ..........................*/

	{ group=nog;

 	 for(i=1; i<nog; i++)
  	 {
   		 F=0; q=i<<M; p=q>>1; p--; q--;
     		for(j=1; j<group; j++)
      		{
			q--;
                         a=((r[p][rows]<<1)^MASK[F]); /* CC*/
                         b=q;  /*CC*/
                         t=x[a][rows]; /*CC*/
        		 x[a][rows] = x[b][rows]; /*CC*/
                         x[b][rows] = t;/*CC*/

		F=1-F;   p -= F;
     		 }
    		 group--;
  	 }
	} /* end of 'if' statement */
 } /* end for rows */
  
bitreversalcolumns();
 
/* Input reordering for the columns */
for (cols=0; cols<N; cols++) {
        M=m;
	eo = M%2; M = m>>1;
	group = nog = 1<<M;
	if(eo==1) M++;

/*..................... M=even/odd ..........................*/

	for(i=0; i<(nog-1); i++)
	 {
  		 F=0; q=i<<M; p=q>>1;
  		 for(j=1; j<group; j++)
   	 	{
    	 		 F=1-F; q++;
                         a=((r[cols][p]<<1)^MASK[F]); /* CC*/
                         b=q;  /*CC*/
                         t=x[cols][a]; /*CC*/
        		 x[cols][a] = x[cols][b]; /*CC*/
                         x[cols][b] = t;/*CC*/
                         p += F;
   		 }
 		  group--;
	 }

	if(eo==0) ; else

/*....................... M=odd ..........................*/

	{ group=nog;

 	 for(i=1; i<nog; i++)
  	 {
   		 F=0; q=i<<M; p=q>>1; p--; q--;
     		for(j=1; j<group; j++)
      		{
			q--;
                         a=((r[cols][p]<<1)^MASK[F]); /* CC*/
                         b=q;  /*CC*/
                         t=x[cols][a]; /*CC*/
        		 x[cols][a] = x[cols][b]; /*CC*/
                         x[cols][b] = t;/*CC*/

			/*swap( ((r[cols][p]<<1)^MASK[F]), q );*/
			F=1-F;   p -= F;
     		 }
    		 group--;
  	 }
	} /* end of 'if' statement */
   } /* end for rows */
 } /* End of input reordering algorithm */

/*****************************************************************/
post_adds()
{
     int step,loops,k,ep,j,i,l,cols,rows;

     /* Do divisions by 2 */
     for (i=0; i<N; i++) 
         for (j=1; j<N; j++) 
             x[i][j] = x[i][j]/2;
      
     for (i=1; i<N; i++) 
          for (j=0; j<N; j++)
             x[i][j] = x[i][j] /2;

      /* Postadditions for the rows */
     for (cols=0; cols<N; cols++) {
        step =N;
	loops = 1;
	for (k=1; k<m; k++)  {
	     step = step>>1;
	     ep = step>>1;
	     loops = loops <<1;
	     for (j=0; j<(step>>1); j++) {
	        l=ep;
                for (i=1; i<loops; i++)  {
                       x[l+step][cols] = x[l+step][cols] - x[l][cols];
                       l =l+step;
	         }
		 ep +=1;
	     }
	 }
      }
      
      /* Postaditions for the columns */
       for (rows=0; rows<N; rows++) {
         step =N;
	 loops = 1;
	   for (k=1; k<m; k++)  {
                step = step>>1;
	        ep = step>>1;
            	loops = loops <<1;
		for (j=0; j<(step>>1); j++) {
	            l=ep;
                    for (i=1; i<loops; i++)  {
                          x[rows][l+step] = x[rows][l+step]-x[rows][l];
                          l =l+step;
	            }
		    ep +=1;
		  }
	     }
       }
     
} /* end of post additions */
/*******************************************************************/
  ans_read()
{
	int rows,cols;
	double value;
	char file_name[50];
	FILE *fi,*fopen();


	printf("Enter name of file to be FCTransformed: ");
	scanf("%s",file_name);
     	if((fi=fopen(file_name,"r"))==NULL) {
		printf("cannot open file  '%s' \n",file_name);
		exit(1);
	}
        for (rows=0; rows<N; rows++) {
             for (cols=0; cols<N; cols++) {
                  /* printf("rows= %d    cols=%d   \n",rows,cols);  */
                   if ((fscanf(fi,"%lf",&value)!=EOF))  {
		        x[cols][rows] = value;
		}
             }
        } 
	fclose(fi);
        printf("OK with reading of file ");
}/* End of "ans_read" function */
/**********************************************************************/
ans_write()
{
	int rows,cols;
	FILE *fo,*fopen();
printf("N end=%d\n",N); 
	fo=fopen("data.out_new","w");
        for (rows=0; rows<N0; rows++) {
            for (cols=0; cols<N0; cols ++) {
               printf("%d  %d  %lf\n",cols,rows,x[cols][rows]); 
                fprintf(fo,"%d  %d  %lf\n",cols,rows,x[cols][rows]);
            }
        }
 	fclose(fo);

	puts("  ");
	puts("Your transformed output file is:  data.out_new");
} /* end of "ans_write" function */
